#ifndef FILE_CIRCULAIRE_H
#define FILE_CIRCULAIRE_H

#include <vector>
#include <iterator>
#include <type_traits>

template <class It>
   class circular_iterator {
      It debut_src, fin_src, debut_circ, cur;
   public:
      using value_type = typename std::iterator_traits<It>::value_type;
      using iterator_category = typename std::iterator_traits<It>::iterator_category;
      using difference_type = typename std::iterator_traits<It>::difference_type;
      using pointer = typename std::iterator_traits<It>::pointer;
      using reference = typename std::iterator_traits<It>::reference;
      circular_iterator(It ds, It fs, It dcirc)
         : debut_src{ ds }, fin_src{ fs }, debut_circ{ dcirc }, cur{ dcirc } {
      }
      circular_iterator& operator++() {
         ++cur;
         if (cur == debut_circ)
            cur = fin_src;
         else if (cur == fin_src)
            if (debut_circ == debut_src)
               cur = fin_src;
            else
               cur = debut_src;
         return *this;
      }
      circular_iterator operator++(int) {
         circular_iterator temp { *this };
         operator++();
         return temp;
      }
      bool operator== (const circular_iterator &it) const noexcept {
         return cur == it.cur;
      }
      bool operator!= (const circular_iterator &it) const noexcept {
         return !(*this == it);
      }
      bool operator== (const It& it) const noexcept {
         return cur == it;
      }
      bool operator!= (const It &it) const noexcept {
         return !(*this == it);
      }
      pointer operator->() {
         return *cur;
      }
      reference operator*() {
         return *cur;
      }
      const reference operator*() const {
         return *cur;
      }
      bool operator<(const circular_iterator &it) const {
         return cur < it.cur;
      }
   };

template <class, class = void>
struct is_iterator : std::false_type {};
template <class T>
struct is_iterator<
   T, std::void_t<decltype(std::declval<std::iterator_traits<T>>())>>
   : std::true_type {
};
template <class T>
static constexpr bool is_iterator_v = is_iterator<T>();

template <class T,
          template <class, class> class C = std::vector,
          class A = std::allocator<T>>
   class file_circulaire {
   public:
      using conteneur_type = C<T,A>;
      using value_type = typename conteneur_type::value_type;
      using size_type = typename conteneur_type::size_type;
      using reference = typename conteneur_type::reference;
      using pointer = typename conteneur_type::pointer;
      using iterator = typename conteneur_type::iterator;
      using const_iterator = typename conteneur_type::const_iterator;
      using reverse_iterator = typename conteneur_type::reverse_iterator;
      using const_reverse_iterator = typename conteneur_type::const_reverse_iterator;
      using circular_iterator = ::circular_iterator<iterator>;
      using const_circular_iterator = ::circular_iterator<const_iterator>;
      using circular_reverse_iterator = ::circular_iterator<reverse_iterator>;
      using const_reverse_circular_iterator = ::circular_iterator<const_reverse_iterator>;
   private:
      conteneur_type elems{};
   public:
      file_circulaire() = default;
      template <class It>
         file_circulaire(It debut, It fin, std::enable_if_t<is_iterator_v<It>, void*> = {}) : elems{ debut, fin } {
         }
      void push_back(const value_type &val) {
         elems.push_back(val);
      }
      void push_back(value_type &&val) {
         elems.push_back(std::move(val));
      }
      size_type size() const noexcept {
         return elems.size();
      }
      iterator begin() noexcept {
         return elems.begin();
      }
      const_iterator begin() const noexcept {
         return elems.begin();
      }
      iterator end() noexcept {
         return elems.end();
      }
      const_iterator end() const noexcept {
         return elems.end();
      }
      reverse_iterator rbegin() noexcept {
         return elems.rbegin();
      }
      const_reverse_iterator rbegin() const noexcept {
         return elems.rbegin();
      }
      reverse_iterator rend() noexcept {
         return elems.rend();
      }
      const_reverse_iterator rend() const noexcept {
         return elems.rend();
      }
      //
      // TODO:
      //
      // voir comment gerer les rbegin_from() et rend_from()
      // 
      circular_iterator begin_from(iterator cur = begin()) {
         return { begin(), end(), cur };
      }
      circular_iterator end_from(iterator cur = begin()) {
         return { begin(), end(), end() };
      }
      const_circular_iterator begin_from(const_iterator cur = begin()) const {
         return { begin(), end(), cur };
      }
      const_circular_iterator end_from(const_iterator cur = begin()) const {
         return { begin(), end(), end() };
      }
      circular_reverse_iterator rbegin_from(reverse_iterator cur = rbegin()) {
         return { rbegin(), rend(), cur };
      }
      circular_reverse_iterator rend_from(reverse_iterator cur = rbegin()) {
         return { rbegin(), rend(), rend() };
      }
      const_reverse_circular_iterator rbegin_from(const_reverse_iterator cur = rbegin()) const {
         return { rbegin(), rend(), cur };
      }
      const_reverse_circular_iterator end_from(const_reverse_iterator cur = rbegin()) const {
         return { rbegin(), rend(), rend() };
      }
   };

/*
   //
   // code de test potentiel...
   //
   file_circulaire<int> f;
   for (int i = 0; i < 5; ++i)
      f.push_back (i);
   std::copy (f.begin (), f.end (),
              std::ostream_iterator <file_circulaire<int>::value_type> (std::cout, " "));
   std::cout << std::endl;
   std::copy (f.begin_from (f.begin () + 2), f.end_from (f.begin () + 2),
              std::ostream_iterator <file_circulaire<int>::value_type> (std::cout, " "));
   std::cout << std::endl;
   std::copy (f.rbegin_from (f.rbegin () + 2), f.rend_from (f.rbegin () + 2),
              std::ostream_iterator <file_circulaire<int>::value_type> (std::cout, " "));
   std::cout << std::endl;
*/

#endif
